/*	maze_type.h (version using stack2.h)
	
	A class which implements an abstract maze type.
	The maze is a rectangle (or box) of blocks with the ability
	to find paths from a start to a finish position within the maze.
	Blocks are a separate type used as part of the template to define
	maze and must at a minimum implement certain features:
		*  defines a direction_t type with circular overloaded ++ operator
		*  a passable(direction_t) member function which returns true
		   if you can move that direction through the block.
		*  a leave_crumb() method to mark the block as having been visited.
		*  a remove_crumb() method to undo leave_crumb.
		*  a crumbed() method which returns true if the block has been visited.
		*  defines a block_t type to store contents
		*  a clear member of type block_t which means "empty"
		*  set(block_t) changes the contents of the block.
		*  contents() returns the value of the block's contents.
		 	
 */
 
#ifndef	ANTHONY_MAZE_TYPE_H
#define	ANTHONY_MAZE_TYPE_H

#include	<iostream>
#include	<iomanip>
#include	<list>
#include	<algorithm>

#include	"stack2.h"								// linked list stack

namespace	anthony_maze {

// selection functions
std::size_t	most_recent( std::size_t upperbound );
std::size_t	oldest( std::size_t upperbound );
std::size_t	random( std::size_t upperbound );
 

template	<class Type>
class	maze_type
{
	public:
		typedef	unsigned int	size_t;

		maze_type( size_t first_dim, size_t second_dim, size_t third_dim = 1);
		~maze_type();
		
		void	find_path( const Type::coord& start, const Type::coord& finish );
		bool	init_maze();
		bool	make_maze();
		
		// Both print_maze() and place_wall() require that a public member called Type::wall 
		// of type Type::block_t exists within Type.
		// Override these functions if different functionality is desired.
		void	print_maze( std::ostream& sout ) const;
		void	place_wall( const Type::coord& where );
		void	clear_wall( const Type::coord& where );
		Type::block_t block_type( const Type::coord& where );
		void	block_type( const Type::coord& where, Type::block_t type );

		// made these public because I am lazy -- 052904
		size_t	xsize;
		size_t	ysize;
		size_t	zsize;
		
	private:
		Type*		mazeblocks;
		
		size_t	element( size_t x, size_t y, size_t z = 1 ) const;
		bool		can_move( Type::coord& position, const Type::direction_t& which_way );
		bool		not_visited( Type::coord& position, const Type::direction_t& which_way );
		void		move( Type::coord& position, const Type::direction_t& which_way );
		bool		is_uncarved( Type::coord& position );
		
};


// constructor & destructor

template	<class Type>
maze_type<Type>::maze_type( size_t first_dim, size_t second_dim, size_t third_dim ) 
{
	// save the size of the maze
	xsize = first_dim;	
	ysize = second_dim;
	zsize = third_dim;
	
	// allocate an array of blocks for the maze
	try	{
	mazeblocks = new	Type[xsize*ysize*zsize];	// allocate a one dimensional array because higher dimensions must be constant expressions
	}
	catch	(bad_alloc)	{
		cerr	<< "Unable to allocate the maze storage." << endl;
		exit(1);
	}
	
	// clear the maze
	for	(int x = 1; x <= xsize; x++)	
		for	(int y = 1; y <= ysize; y++)
			for	(int z = 1; z <= zsize; z++)
				mazeblocks[element(x,y,z)].set( Type::clear );

}

template	<class Type>
maze_type<Type>::~maze_type()
{
	delete []	mazeblocks;
}


// member functions

template	<class Type>
void	maze_type<Type>::find_path( const Type::coord& start, const Type::coord& finish )
{
	// make sure that start and finish are not walls
	if	( mazeblocks[element(start.first,start.second)].contents() == Type::wall )	{
		cout	<< "The start position is occupied by a wall!" << endl;
		return;
	}
	if	( mazeblocks[element(finish.first,finish.second)].contents() == Type::wall )	{
		cout	<< "The finish position is occupied by a wall!" << endl;
		return;
	}

	anthony_stack2::stack<coord> 	path; 
	anthony_stack2::stack<coord>	reverse;
	Type::coord		position;					// current position
	Type::direction_t temp;						// temporary direction holder
	bool			path_found;
	bool			no_path;					// true if no path exists
	
	no_path = false;
	path_found = false;
	position = start;
	if	( ! path.push(position) )	{			// push the starting coordinates
		cerr	<< "Ran out of memory!" << endl;
		return;
	}
	
	// mark start as "visited"
	mazeblocks[element(position.first, position.second)].leave_crumb();
	
	// check to see if start and finish are the same
	if	( start.first == finish.first && start.second == finish.second )	{
		path_found = true; 
		if	( ! path.push(finish) )	{			// push finish so it will print out
			cerr	<< "Ran out of memory!" << endl;
			return;
		}
	}

	// continue searching for a path until finish is found or all possibilities are exhausted
	while	( ! path_found && ! no_path )	{
		if	( can_move(position, direction) && not_visited(position, direction) )	{
			move(position, direction);
			if	( ! path.push(position) )	{	// push new position onto stack
				cerr	<< "Ran out of memory!" << endl;
				return;
			}
			// check to see if we have reached the finish
			if	( position.first == finish.first && position.second == finish.second )
				path_found = true;
			// mark as visited
			mazeblocks[element(position.first, position.second)].leave_crumb();
			direction = Type::first_dir;			// reset to first direction
		}
		else	if	( direction != Type::last_dir )	// still more directions to try
				++direction;
		else	{
			path.pop(position);
			path.pop(position);				// backup one block
			if	( path.empty() )	{			// back at the start position
				temp = Type::first_dir;
				no_path = true; 				// assume true for the moment
				do	{					// test each direction to see if any options left
					if	( can_move(position, temp) && not_visited(position, temp) )	{
						no_path = false;		// temp direction is unexplored
						break;
					}
					else	++temp;
				}
				while	( temp != Type::first_dir );
			}
			path.push(position);				// push current position
			direction = Type::first_dir;			// reset to first direction
		}
	}
	
	// print the path out if one was found
	if	(path_found)	{
		cout	<< "Path found: " << endl;
		while	( !path.empty() )	{
			path.pop(position);
			if	( ! reverse.push(position) )	{	// reverse order of coordinates in path
				cerr	<< "Ran out of memory!" << endl;
				return;
			}
		}
		
		reverse.pop(position);					/// pop the starting position
		cout	<< setw(4) << position.first << setw(4) << position.second << "   Start";
		while	( !reverse.empty() )	{
			reverse.pop(position);
			cout	<< endl << setw(4) << position.first << setw(4) << position.second;
		}
		cout	<< "   Finish" << endl;
	}
	else	cout	<< "No path found from " << start.first << "  " << start.second << " to " << finish.first << "  " << finish.second << endl;
	 
	return;
}

template	<class Type>
bool	maze_type<Type>::init_maze()
{
	if	( (xsize%2) == 0 or (ysize%2) == 0 )	{
		cerr << "Maze dimensions cannot be even when generating a random maze." << endl;
		return false;
	}
	
	// initialize the maze as a grid of enclosed cells
	// even rows and columns are cells while odd ones are walls
	bool	makeWall = false;
	for	(int x = 1; x <= xsize; x++)	{	
		for	(int y = 1; y <= ysize; y++)	{
			if	( x == 1 or x == xsize or y == 1 or y == ysize )	makeWall = true;
			else	if	( (x%2) == 1 or (y%2) == 1 )				makeWall = true;
			else	makeWall = false;
			mazeblocks[element(x,y)].set( makeWall ? Type::wall : Type::clear );
		}
	}
	
	return true;
}

typedef	std::size_t (*selection_func_t)(std::size_t);

template	<class Type>
bool	maze_type<Type>::make_maze()
{
	if	(! init_maze() )	return false;
	
	typedef	list<Type::coord>	CoordList;
	
	int	width  = xsize / 2;
	int	height = ysize / 2;
	Type::coord	curcell, startcell;	// numbering cell positions sequentially (i.e. ignoring walls)
	CoordList carved_cells;	// cells which have been carved into from most recent to oldest
	int	index, remaining;
	selection_func_t	select = most_recent;
	
	// choose where to put the entrance
	bool	onleft = rand() % 2;
	if	(onleft)	{	// either leftmost column
		startcell.first = 1;
		startcell.second = rand()%height + 1;
	}	
	else	{			// or bottom row
		startcell.first = rand()%width + 1;
		startcell.second = height;
	}
	
	// "carve" out the entrance and add interior cell to start the process
	if	(onleft)	mazeblocks[element(startcell.first*2-1, startcell.second*2)].set( Type::start );
	else			mazeblocks[element(startcell.first*2, startcell.second*2+1)].set( Type::start );
	carved_cells.push_front(startcell);
	
	/* select carved cells, one at a time, and carve to one of its uncarved neighbors */
	while	( remaining = carved_cells.size() )	{
		index = select(remaining);
		// find the index-th element of the list carved_cells
		CoordList::iterator iter;
		int i;
		for	( i = 0, iter = carved_cells.begin(); (i<index) and (iter != carved_cells.end()); ++i, ++iter);
		if	(iter == carved_cells.end())	return false;	// this shouldn't happen
		curcell = *iter;
		
		// find all uncarved neighbor cells
		bool		uncarved[4] = {false};				// north, east, south, west
		int		uncarvedNeighbors = 0;
		Type::coord	neighbor;
		
		if	( curcell.second != 1 )	{				// skip a check if on edge of maze
			neighbor.first = curcell.first*2; neighbor.second = (curcell.second-1)*2;	// north
			if	( uncarved[0] = is_uncarved(neighbor) )	++uncarvedNeighbors;
		}
		if	( curcell.first != width )	{
			neighbor.first = (curcell.first+1)*2; neighbor.second = curcell.second*2;	// east
			if	( uncarved[1] = is_uncarved(neighbor) )	++uncarvedNeighbors;
		}
		if	( curcell.second != height )	{
			neighbor.first = curcell.first*2; neighbor.second = (curcell.second+1)*2;	// south
			if	( uncarved[2] = is_uncarved(neighbor) )	++uncarvedNeighbors;
		}
		if	( curcell.first != 1 )	{
			neighbor.first = (curcell.first-1)*2; neighbor.second = curcell.second*2;	// west
			if	( uncarved[3] = is_uncarved(neighbor) )	++uncarvedNeighbors;
		}
	
		if	( uncarvedNeighbors == 0 )	{			// nowhere to go
			carved_cells.erase(iter);				// remove cell from further consideration
		}
		else	{
			// choose at random one of the uncarved neighbors
			int	celldirection;
			int	choice = rand() % uncarvedNeighbors;
			for	( int j = 0, k = 0; j <= choice; (++k) %= 4 )	{
				if	(uncarved[k])	{
					celldirection = k;
					j++;
				}
			}
			// remove the wall between the current cell and the chosen neighbor
			switch (celldirection)	{
				case	0:	// north
					mazeblocks[element(curcell.first*2, curcell.second*2-1)].set( Type::clear );
					neighbor.first = curcell.first; neighbor.second = curcell.second - 1;
					break;
				case	1:	// east
					mazeblocks[element(curcell.first*2+1, curcell.second*2)].set( Type::clear );
					neighbor.first = curcell.first + 1; neighbor.second = curcell.second;
					break;
				case	2:	// south
					mazeblocks[element(curcell.first*2, curcell.second*2+1)].set( Type::clear );
					neighbor.first = curcell.first; neighbor.second = curcell.second + 1;
					break;
				case	3:	// west
					mazeblocks[element(curcell.first*2-1, curcell.second*2)].set( Type::clear );
					neighbor.first = curcell.first - 1; neighbor.second = curcell.second;
					break;
				default:
					cerr	<< "Invalid direction at line " << __LINE__
						<< " in file " << __FILE__ << endl;
					return false;
					break;
			}
			
			// add newly carved cell to list
			carved_cells.push_front(neighbor);
			//cout	<< endl;
			//print_maze(cout);
		}
	}
	
	return true;
}

template	<class Type>
void	maze_type<Type>::print_maze( std::ostream& sout ) const
{
	for	(int y = 1; y <= ysize; y++)	{
		sout	<< "\t";
		for	(int x = 1; x <= xsize; x++)
			switch ( mazeblocks[element(x,y)].contents()  )	{
				case Type::wall:	sout	<< "X "; break;
				case Type::start:	sout	<< "S "; break;
				case Type::exit:	sout	<< "E "; break;
				case Type::gold:	sout	<< "G "; break;
				default:		sout	<< "  "; break;
			}
		sout	<< endl;
	}
	
	return;
}

template	<class Type>
void	maze_type<Type>::place_wall( const Type::coord& where )
{
	// cout << "place_wall() at coordinates " << where.first << "   " << where.second << endl;
	mazeblocks[element(where.first, where.second)].set( Type::wall );
	return;
}

template	<class Type>
void	maze_type<Type>::clear_wall( const Type::coord& where )
{
	// cout << "clear_wall() at coordinates " << where.first << "   " << where.second << endl;
	mazeblocks[element(where.first, where.second)].set( Type::clear );
	return;
}

template	<class Type>
Type::block_t maze_type<Type>::block_type( const Type::coord& where )
{
	return mazeblocks[element(where.first, where.second)].contents();
}

template	<class Type>
void	maze_type<Type>::block_type( const Type::coord& where, Type::block_t type )
{
	mazeblocks[element(where.first, where.second)].set(type);
	return;
}

/* element()
   Returns index within mazeblocks to the square (x,y,z) of the maze.
   Precondition: 1 <= x <= xsize, 1 <= y <= ysize, 1 <= z <= zsize
 */
template	<class Type>
maze_type<Type>::size_t maze_type<Type>::element( size_t x, size_t y, size_t z ) const
{
	size_t element_num;

	// all indexes start at 1 !
	element_num = ( (z-1)*ysize*xsize ) + ( (y-1)*xsize ) + (x-1);
	
	return element_num;
}

/* can_move()
   Determines whether it is possible to move in a given direction from a certain position.
 */
template	<class Type>
bool	maze_type<Type>::can_move( Type::coord& pos, const Type::direction_t& which_way )
{
	switch (which_way)	{
		case	Type::north:
			if	( pos.second == 1 ) 	return false;	// cannot move up from top row
			if	( mazeblocks[element(pos.first, pos.second-1)].passable(which_way) )
				return true;
			break;
			
		case	Type::east:
			if	( pos.first == xsize )	return false;	// cannot move right from last column
			if	( mazeblocks[element(pos.first+1, pos.second)].passable(which_way) )
				return true;
			break;
			
		case	Type::south:
			if	( pos.second == ysize )	return false;	// cannot move down from bottom row
			if	( mazeblocks[element(pos.first, pos.second+1)].passable(which_way) )
				return true;
			break;
			
		case	Type::west:
			if	( pos.first == 1 ) 	return false;	// cannot move left from first column
			if	( mazeblocks[element(pos.first-1, pos.second)].passable(which_way) )
				return true;
			break;
			
		default:
			return false;
	}

	return false;
}

/* move()
   Actually moves in the specified direction from the given position and returns the new position.
 */
template	<class Type>
void	maze_type<Type>::move( Type::coord& pos, const Type::direction_t& which_way )
{
	switch (which_way)	{
		case	Type::north:	
			--pos.second;					// move north
			break;
			
		case	Type::east:
			++pos.first;					// move east
			break;
			
		case	Type::south:
			++pos.second;					// move south
			break;
			
		case	Type::west:
			--pos.first;					// move west
			break;
			
		default:
			break;
	}

	return;
}

/* not_visited()
   Determines whether the next block in direction which_way from coordinates pos has already been visited.
 */
template	<class Type>
bool	maze_type<Type>::not_visited( Type::coord& pos, const Type::direction_t& which_way )
{
	switch (which_way)	{
		case	Type::north:
			if	( mazeblocks[element(pos.first, pos.second-1)].crumbed() )
				return false;
			break;
			
		case	Type::east:
			if	( mazeblocks[element(pos.first+1, pos.second)].crumbed() )
				return false;
			break;
			
		case	Type::south:
			if	( mazeblocks[element(pos.first, pos.second+1)].crumbed() )
				return false;
			break;
			
		case	Type::west:
			if	( mazeblocks[element(pos.first-1, pos.second)].crumbed() )
				return false;
			break;
			
		default:
			return false;
	}

	return true;
}

/* is_uncarved()
   We consider a cell to be uncarved if it is surrounded in all directions by walls.
 */
template	<class Type>
bool	maze_type<Type>::is_uncarved( Type::coord& position )
{
	Type::direction_t	direction, last;
	
	direction = Type::first_dir;
	do	{
		if	( can_move(position, direction) )	return false;
		last = direction;
		++direction;
	}
	while	( last != Type::last_dir );	// still more directions to try
	
	return true;
}

}	// closure of namespace

#endif
